"""
   Copyright 2017 Cédric BOUYSSET

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

from rdkit import Chem
import os.path
from .residue import *
from .utils import *

class Protein:
    """Class for a protein"""

    def __init__(self, inputFile, reference, cutoff, residueList):
        """Initialization of the protein, defined by a list of residues"""
        self.residueList = residueList
        self.residues = {}
        self.inputFile = inputFile
        fileExtension = os.path.splitext(inputFile)[1]
        if fileExtension == '.mol2':
            self.residuesFromMOL2File()
        else:
            raise ValueError('{} files are not supported for the protein.'.format(fileExtension[1:].upper()))
        if self.residueList == None:
            self.residueList = self.detectCloseResidues(reference, cutoff)
        self.cleanResidues()

    def __repr__(self):
        return self.inputFile

    def residuesFromMOL2File(self):
        """Read a MOL2 file and assign each line to an object of class Atom"""
        # Create a molecule with RDKIT
        self.mol = Chem.MolFromMol2File(self.inputFile, sanitize=True, removeHs=False)
        # Read the atoms directly from the MOL2 file
        rec = mol2_reader(self.inputFile, ignoreH=False)
        # Loop through each RDKIT atom and assign them to a residue
        residues = {}
        coordinates = self.mol.GetConformer().GetPositions()
        for i,(rd_at, at) in enumerate(zip(self.mol.GetAtoms(), rec)):
            residue = at['residue']
            if residue not in residues:
                residues[residue] = []
            atom = {
                'resid'  :     residue,
                'resname':     residue[:3],
                'chain'  :     at['chain'],
                'coordinates': coordinates[i],
                'charge':      float(rd_at.GetProp('_TriposPartialCharge')),
                'aromatic':    rd_at.GetIsAromatic(),
                'ring':        rd_at.IsInRing(),
            }
            residues[residue].append(atom)
        # Create residues objects
        for residue in residues:
            self.residues[residue] = Residue(residues[residue])

    def detectCloseResidues(self, reference, cutoff):
        """Detect residues close to a reference ligand"""
        residueList = []
        for residue in self.residues:
            d = euclidianDistance(reference.centroid, self.residues[residue].centroid)
            if d <= cutoff:
                residueList.append(self.residues[residue].resid)
        return residueList

    def cleanResidues(self):
        """Cleans the residues of the protein to only keep those in self.residueList"""
        residues = {}
        for residue in self.residues:
            if residue in self.residueList:
                residues[residue] = self.residues[residue]
        self.residues = residues



#!/usr/bin/python3

"""
   Copyright 2017 Cédric BOUYSSET

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

from .ligand import *
from .protein import *
from .fingerprint import *

def main(args):

    # Read files
    fingerprint = Fingerprint(args.json, args.interactions)
    if args.verbose:
        print('Built fingerprint generator using the following bitstring:')
        print(fingerprint)
    reference   = Ligand(args.reference)
    if args.verbose:
        print('Read reference molecule from ' + str(reference))
    protein     = Protein(args.protein, reference, cutoff=args.cutoff, residueList=args.residues)
    if args.verbose:
        print('Read protein from ' + str(protein))
    length      = len(args.interactions)
    # Print residues on terminal:
    for residue in protein.residues:
        print('{resid: >{length}s}'.format(resid=protein.residues[residue].resid, length=length), end='')
    print()
    # Generate the IFP between the reference ligand and the protein
    fingerprint.generateIFP(reference, protein)
    print(reference.IFP)
    # Loop over ligands:
    ligandList = []
    for lig in args.ligand:
        ligand = Ligand(lig)
        # Generate the IFP between a ligand and the protein
        fingerprint.generateIFP(ligand, protein)
        # Calculate similarity
        score = ligand.getSimilarity(reference, args.score, args.alpha, args.beta)
        ligand.setSimilarity(score)
        print(ligand.IFP, '{:.4f}'.format(ligand.score))
        ligandList.append(ligand)
    # Output
    if args.output:
        with open(args.output, 'w') as file:
            file.write('File,SimilarityScore')
            for residue in protein.residues:
                file.write(',{}'.format(protein.residues[residue].resid))
            file.write('\n')
            CSIFP = ','.join(reference.IFP[i:i+length] for i in range(0, len(reference.IFP), length))
            file.write('{},1.0000,{}\n'.format(args.reference, CSIFP))
            for ligand in ligandList:
                CSIFP = ','.join(ligand.IFP[i:i+length] for i in range(0, len(ligand.IFP), length))
                file.write('{},{:.4f},{}\n'.format(ligand, ligand.score, CSIFP))

if __name__ == '__main__':
    main(args)



"""
   Copyright 2017 Cédric BOUYSSET

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

from rdkit import Chem
from .utils import getCentroid

class Residue:
    """Class for a residue in a protein"""

    def __init__(self, atoms):
        self.atoms    = atoms
        self.resname  = atoms[0]['resname'] # 3 letter code
        self.resid    = atoms[0]['resid']   # unique identifier for the residue
        self.centroid = getCentroid(self.atoms)

    def __repr__(self):
        return self.resid



"""
   Copyright 2017 Cédric BOUYSSET

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

from math import sqrt
import re

def euclidianDistance(a,b):
    """Euclidian distance between 2 lists of coordinates"""
    return sqrt(sum([(xa-xb)**2 for xa,xb in zip(a,b)]))

def getCentroid(atoms):
    """Centroid coordinates (XYZ) of a list of atoms"""
    return [sum([atom['coordinates'][i] for atom in atoms])/len(atoms) for i in range(3)]

def mol2_reader(mol2_file, ignoreH):
    '''A simple MOL2 file reader. Can only read files with a single molecule in it.
    Returns a molecules. Each molecule is a list of atoms, where an atom
    is a dictionnary containing atom informations.'''
    num_atoms_lines = []
    first_lines     = []

    # Read file
    with open(mol2_file, "r") as f:
        lines = f.readlines()

    # Search for the line where the number of atoms is, and the first line where atom coordinates are readable
    for i, line in enumerate(lines):
        search_molecule = re.search(r'@<TRIPOS>MOLECULE', line)
        search_atom     = re.search(r'@<TRIPOS>ATOM', line)

        if search_molecule:
            # line with the number of atoms
            num_atoms_lines.append(i+2)
        elif search_atom:
            # first line with atom coordinates
            first_lines.append(i+1)

    for num_atoms_line, first_line in zip(num_atoms_lines, first_lines):
        molecule = get_mol_from_mol2(num_atoms_line, first_line, lines, ignoreH)

    return molecule

def get_mol_from_mol2(num_atoms_line, first_line, lines, ignoreH):
    '''Extracts a molecule from a mol2 file.
    num_atoms_line: index of the line containing the number of atoms, bonds...etc.
    first_line: index of the first line of the molecule to be extracted
    Returns a molecule as a list of atoms. An atom is a dictionnary containing
    atom informations that aren't read by the RDkit reader: index, chain, residue'''
    # Read number of atoms directly from the corresponding line
    data      = lines[num_atoms_line].split()
    num_atoms = int(data[0])

    # create molecule containing a list of atoms
    MOL = []
    # Fill the list with atomic parameters and coordinates
    for line in range(first_line, first_line + num_atoms):
        data = lines[line].split()
        if ignoreH: # if ignore H
            if data[5] == 'H': # if the atom read is an H
                continue # skip this atom
        # if it's not an H, or we don't ignore H, add this atom
        MOL.append(
        {
            'chain'   : str(data[6]),
            'residue' : str(data[7]),
        })
    return MOL



import argparse, textwrap, os, sys
from . import prolif
from .version import __version__

class Range:
    """Class to raise an exception if the value is not between start and end.
    Used for alpha and beta in Tversky"""
    def __init__(self, start, end):
        self.start = start
        self.end = end
    def __eq__(self, other):
        return self.start <= other <= self.end
    def __repr__(self):
        return '{} to {}'.format(self.start, self.end)

def cli():
    jsonpath = os.path.join(os.path.dirname(__file__),'prolif.json')
    description = 'ProLIF: Protein Ligand Interaction Fingerprints\nGenerates Interaction FingerPrints (IFP) and a similarity score for protein-ligand interactions'
    epilog = 'Mandatory arguments: --reference --ligand --protein\nMOL2 files only.'
    # Argparse
    parser = argparse.ArgumentParser(description=description, epilog=epilog,
        formatter_class=argparse.RawTextHelpFormatter)

    group_input = parser.add_argument_group('INPUT arguments')
    group_input.add_argument("-r", "--reference", metavar='fileName', type=str, required=True,
        help="Path to your reference ligand.")
    group_input.add_argument("-l", "--ligand", metavar='fileName', type=str, nargs='+', required=True,
        help="Path to your ligand(s).")
    group_input.add_argument("-p", "--protein", metavar='fileName', type=str, required=True,
        help="Path to your protein.")
    group_input.add_argument("--residues", type=str, nargs='+', default=None,
        help="Residues chosen for the interactions. Default: automatically detect residues within --cutoff of the reference ligand")
    group_input.add_argument("--cutoff", metavar='float', type=float, required=False, default=5.0,
        help="Cutoff for automatic residue detection. Default: 5.0 angströms")
    group_input.add_argument("--json", metavar='fileName', type=str, default=jsonpath,
        help="Path to a custom parameters file. Default: prolif.json")

    group_output = parser.add_argument_group('OUTPUT arguments')
    group_output.add_argument("-o", "--output", metavar='filename', type=str,
        help="Path to the output CSV file")
    group_output.add_argument("-v", "--verbose", action="store_true", help="Increase terminal output verbosity")
    group_output.add_argument("--version", action="version",
        version='ProLIF {}'.format(__version__) ,help="Show version and exit")

    group_args = parser.add_argument_group('Other arguments')
    group_args.add_argument("--interactions", metavar="bit", nargs='+',
        choices=['HBdonor','HBacceptor','cation','anion','FaceToFace','FaceToEdge',
            'pi-cation','hydrophobic','XBdonor','metal'],
        default=['HBdonor','HBacceptor','cation','anion','FaceToFace','FaceToEdge','pi-cation','hydrophobic'],
        help=textwrap.dedent("""List of interactions used to build the fingerprint.
    -) hydrogen bond: HBdonor, HBacceptor
    -) halogen bond:  XBdonor
    -) ionic: cation, anion
    -) pi-stacking: FaceToFace, FaceToEdge
    -) hydrophobic
    -) pi-cation
    -) metal
    Default: HBdonor HBacceptor cation anion FaceToFace FaceToEdge pi-cation hydrophobic"""))
    group_args.add_argument("--score", choices=['tanimoto', 'dice', 'tversky'], default='tanimoto',
        help=textwrap.dedent("""Similarity score between molecule A and B :
    Let 'a' and 'b' be the number of bits activated in molecules A and B, and 'c' the number of activated bits in common.
    -) tanimoto : c/(a+b-c). Used by default
    -) dice     : 2c/(a+b)
    -) tversky  : c/(alpha*(a-c)+beta*(b-c)+c)
    """))
    group_args.add_argument("--alpha", metavar="int", type=float, choices=[Range(0.0, 1.0)], default=0.7,
        help="Alpha parameter for Tversky. Default: 0.7")
    group_args.add_argument("--beta", metavar="int", type=float, choices=[Range(0.0, 1.0)], default=0.3,
        help="Beta parameter for Tversky. Default: 0.3")

    # Parse arguments from command line
    args = parser.parse_args()

    prolif.main(args)



"""
   Copyright 2017 Cédric BOUYSSET

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

import json
from rdkit import DataStructs
from random import randint

class Fingerprint:
    """Class that generates an interaction fingerprint between a protein and a ligand"""

    def __init__(self, json_file, interactions):
        # read parameters from json file
        with open(json_file) as data_file:
            self.prm = json.load(data_file)
        # read interactions to compute
        self.interactions = interactions

    def __repr__(self):
        return ' '.join(self.interactions)

    def hasHydrophobic(self, ligand, residue):
        """Get the presence or absence of an hydrophobic interaction between
        a residue and a ligand"""
        return randint(0, 1)

    def hasFaceToFace(self, ligand, residue):
        """Get the presence or absence of an aromatic face to face interaction
        between a residue and a ligand"""
        return randint(0, 1)

    def hasFaceToEdge(self, ligand, residue):
        """Get the presence or absence of an aromatic face to edge interaction
        between a residue and a ligand"""
        return randint(0, 1)

    def hasHBdonor(self, ligand, residue):
        """Get the presence or absence of a H-bond interaction between
        a residue as an acceptor and a ligand as a donor"""
        return randint(0, 1)

    def hasHBacceptor(self, ligand, residue):
        """Get the presence or absence of a H-bond interaction between
        a residue as a donor and a ligand as an acceptor"""
        return randint(0, 1)

    def hasXBdonor(self, ligand, residue):
        """Get the presence or absence of a Halogen Bond"""
        return randint(0, 1)

    def hasCationic(self, ligand, residue):
        """Get the presence or absence of an ionic interaction between a residue and a ligand as a cation"""
        return randint(0, 1)

    def hasAnionic(self, ligand, residue):
        """Get the presence or absence of an ionic interaction between a residue and a ligand as an anion"""
        return randint(0, 1)

    def hasPiCation(self, ligand, residue):
        """Get the presence or absence of Pi-Cation interaction"""
        return randint(0, 1)

    def hasMetal(self, ligand, residue):
        """Get the presence or absence of a metal complexation"""
        return randint(0, 1)

    def generateBitstring(self, ligand, residue):
        """Generate the complete bitstring for the interactions of a residue with a ligand"""
        bitstring = []
        if 'HBdonor' in self.interactions:
            bitstring.append(self.hasHBdonor(ligand, residue))
        if 'HBacceptor' in self.interactions:
            bitstring.append(self.hasHBacceptor(ligand, residue))
        if 'cation' in self.interactions:
            bitstring.append(self.hasCationic(ligand, residue))
        if 'anion' in self.interactions:
            bitstring.append(self.hasAnionic(ligand, residue))
        if 'FaceToFace' in self.interactions:
            bitstring.append(self.hasFaceToFace(ligand, residue))
        if 'FaceToEdge' in self.interactions:
            bitstring.append(self.hasFaceToEdge(ligand, residue))
        if 'pi-cation' in self.interactions:
            bitstring.append(self.hasPiCation(ligand, residue))
        if 'hydrophobic' in self.interactions:
            bitstring.append(self.hasHydrophobic(ligand, residue))
        if 'XBdonor' in self.interactions:
            bitstring.append(self.hasXBdonor(ligand, residue))
        if 'metal' in self.interactions:
            bitstring.append(self.hasMetal(ligand, residue))
        return ''.join(str(bit) for bit in bitstring)

    def generateIFP(self, ligand, protein):
        """Generates the complete IFP from each residue's bitstring"""
        IFPvector = DataStructs.ExplicitBitVect(len(self.interactions)*len(protein.residues))
        i=0
        IFP = ''
        for residue in protein.residues:
            bitstring = self.generateBitstring(ligand, protein.residues[residue])
            for value in bitstring:
                if value == '1':
                    IFPvector.SetBit(i)
                i+=1
            IFP += bitstring
        ligand.setIFP(IFP, IFPvector)



from setuptools import setup
from os import path
from codecs import open

here = path.abspath(path.dirname(__file__))

# get version from version.py
__version__ = None
exec(open('prolif/version.py').read())

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='prolif',
      version=__version__,
      description='Protein-Ligand Interaction Fingerprints',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/cbouy/ProLIF',
      author='Cédric Bouysset',
      author_email='bouysset.cedric@gmail.com',
      license='Apache License, Version 2.0',
      classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Operating System :: OS Independent',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Chemistry',
      ],
      keywords='science chemistry biology drug-design chemoinformatics virtual-screening',
      packages=['prolif'],
      entry_points = {
        'console_scripts': ['prolif=prolif.command_line:cli'],
        },
      dependency_links=['git+https://github.com/rdkit/rdkit'],
      test_suite="prolif.tests",
      include_package_data=True,
      zip_safe=False)



"""
   Copyright 2017 Cédric BOUYSSET

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
"""

from rdkit import Chem, DataStructs
import os.path

class Ligand:
    """Class for a ligand"""
    def __init__(self, inputFile):
        """Initialize the ligand from a file"""
        self.inputFile = inputFile
        fileExtension = os.path.splitext(inputFile)[1]
        if fileExtension == '.mol2':
            self.mol = Chem.MolFromMol2File(inputFile, sanitize=True, removeHs=False)
        else:
            raise ValueError('{} files are not supported for the ligand.'.format(fileExtension[1:].upper()))
        # Set Centroid
        coordinates = self.mol.GetConformer().GetPositions()
        self.centroid = [sum([atom[i] for atom in coordinates])/len(coordinates) for i in range(3)]

    def __repr__(self):
        return self.inputFile

    def setIFP(self, IFP, vector):
        """Set the IFP for the ligand, as a bitstring and vector"""
        self.IFP = IFP
        self.IFPvector = vector

    def getSimilarity(self, reference, method, alpha, beta):
        if   method == 'tanimoto':
            return DataStructs.TanimotoSimilarity(reference.IFPvector, self.IFPvector)
        elif method == 'dice':
            return DataStructs.DiceSimilarity(reference.IFPvector, self.IFPvector)
        elif method == 'tversky':
            return DataStructs.TverskySimilarity(reference.IFPvector, self.IFPvector, alpha, beta)

    def setSimilarity(self, score):
        """Set the value for the similarity score between the ligand and a reference"""
        self.score = score
